// Visualization of Bitcoin's overflow bug in 2010
// compile via:  g++ -Wall bitcoin.cpp
#include <climits>
#include <string> 
#include <iostream>
#include <sstream> 
#include <cmath>
#include <iomanip>

  void printUInt(std::string txt, unsigned long long nr) {
      std::cout << std::setw(29) << txt 
                << std::setw(20) << nr << std::endl;
  }
  void printSInt(std::string txt, long long nr) {
      std::cout << std::setw(29) << txt 
                << std::setw(20) << nr << std::endl;
  }

int main (int argc, char *argv[]) {
  std::string delimiter(25, '-');
  std::cout << delimiter << delimiter << std::endl;
  std::cout << "check spec. values (unsigned/signed long long)"
            << std::endl;
  std::cout << delimiter << delimiter << std::endl;
  
  printUInt("ULLONG_MAX : ", ULLONG_MAX);
  unsigned long long twoPow63 = pow(2,63);
  printUInt("2^63 : ", twoPow63);
  // out1: literal interpreted as unsigned!
  unsigned long long out1 = 9223372036854275808; 
  printUInt("out1 (literal) : ", out1);
  // out2: read in from hex code; identical to literal
  unsigned long long out2; 
  std::stringstream ss; ss << std::hex << "7ffffffffff85ee0";
  ss >> out2;
  printUInt("out2 (conversion from hex) : ", out2);
  long long dirSum = static_cast<signed long long>(out1 + out2);
  printSInt("cast (out1+out2) to signed : ", dirSum);
  
  // imitate bitcoin formula
  std::cout << delimiter << delimiter << std::endl;
  std::cout << "imitate original bitcoin formula" << std::endl;
  std::cout << delimiter << delimiter << std::endl;
  // outgoing numbers
  long long nValueOut = 0;
  printSInt("nValueOut = 0 : ", nValueOut);
  nValueOut += out1;
  printSInt("nValueOut += out1 : ", nValueOut);
  nValueOut += out2;
  printSInt("nValueOut += out2 : ", nValueOut);
  nValueOut += 5051000000;
  printSInt("nValueOut += 5051000000 : ", nValueOut);
  
  // incoming numbers
  long long nValueIn = 0;
  printSInt("nValueIn : ", nValueIn);
  long long mineReward = 5000000000;
  printSInt("mineReward : ", mineReward);
  long long investment =   50000000;
  printSInt("investment : ", investment);
  nValueIn += mineReward;
  nValueIn += investment;
  printSInt("nValueIn : ", nValueIn);
  
  // summing all incoming and outgoing numbers
  printSInt("nValueIn - nValueOut : ", nValueIn - nValueOut);
  
  return 0;
}
